# Copyright (C) 2006 Mortal Coil Games
# See LICENSE for details.
"""
Class for an asset client interface 
Handles retrieval and storage of assets.

"""


from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.application.service import Service

from zope.interface import implements

from mv3d.net.client import ServiceLoc
from mv3d.util.conductor import IAssetClient, IPlayerClient


class AssetService(Service):
    """
    You need to add a docstring here ! 
    """
    implements(IAssetClient)

    def __init__(self):
        self.assetcache = {}
        self.downloading = {}
        self.assetServices = []


    def configure(self, nm, cf):
        """
        Configure this client with the config file
        """
        if cf.has_option(nm, "assetServices"):
            asv = cf.get(nm, "assetServices").split(",")
            for a in asv:
                self.assetServices.append(ServiceLoc(a))


    def acquireAssets(self, aids, **kw):
        """ 
        Takes a list of asset ids to acquire
        eventually returns a dict of id:asset
        """
        d = []
        for a in aids:
            d.append(self.acquireAsset(a, **kw))
        def gotAssets(assets):
            d = []
            for a in aids:
                d.append(self.getAsset(a))
            return defer.gatherResults(d)
        return defer.gatherResults(d).addCallback(gotAssets)


    def acquireAsset(self, aid, **kw):
        """
        If we locally have this asset in our
        cache, call its acquire function and return
        the result. Else find the asset and do that.
        """
        if self.downloading.has_key(aid):
            d = defer.Deferred()
            self.downloading[aid].append(d)
            return d
        self.downloading[aid] = []
        d = self.getAsset(aid)
        def gotAsset(a):
            d = defer.maybeDeferred(a.acquireAsset, **kw)
            def done(x):
                for d in self.downloading[aid]:
                    d.callback(a)
                del self.downloading[aid]
                return a
            return d.addCallback(done)
        return d.addCallback(gotAsset)


    @inlineCallbacks
    def getAsset(self, aid):
        """
        Get an asset. If it's in our cache, then
        return it. Otherwise, search with one of our asset
        servers.
        """
        if self.assetcache.has_key(aid):
            returnValue(self.assetcache[aid])
        if not len(self.assetServices):
            p = self.parent.getLocalService(IPlayerClient)
            svcs = yield p.getAssetServices()
            self.assetServices += svcs
        i = yield self.parent.getOneService(self.assetServices)
        a = yield i.getAsset(aid)
        a.parent = self
        self.assetcache[a.getID()] = a
        returnValue(a)


    def isLocal(self):
        """
        We are local so return true
        """
        return True