# Copyright (C) 2006 Mortal Coil Games
# See LICENSE for details.
""" Class for a simulation client interface 

  Handles a player view along with client side physics, graphics, 
  and sound?
  
"""

from twisted.internet.task import coiterate, LoopingCall
from twisted.application.service import Service

from zope.interface import implements

from mv3d.util.conductor import ISimulationClient, IPlayerClient



class ItemIterator:
    """
    Iterates through a list of items
    """
    def __init__(self, parent, items):
        self.parent = parent
        self.items = items.__iter__()


    def next(self):
        """
        implimenting the iter interface
        """
        itm = self.items.next()
        try:
            wasEnabled = self.parent.enabledState.get(itm.getID(),
                itm.isEnabled())
            r = itm.iterate(0.05)
            self.parent.enabledState[itm.getID()] = itm.isEnabled()
            if itm.getID() == self.parent.pcid:
                # print time.time(), itm.isEnabled(), tuple(itm.body.getLinearVelocity())
                for v in self.parent.views.values():
                    v.updateCamera()
            if wasEnabled != itm.isEnabled():
                if wasEnabled:
                    # print "sim disabling", itm
                    try:
                        self.parent.enabled.remove(itm)
                    except ValueError:
                        pass
                    if not itm in self.parent.disabled:
                        self.parent.disabled.append(itm)
                else:
                    # print "sim enabling", itm
                    try:
                        self.parent.disabled.remove(itm)
                    except ValueError:
                        pass
                    if not itm in self.parent.enabled:
                        self.parent.enabled.append(itm)
        except:
            import traceback
            traceback.print_exc()



class SimService(Service):
    """
    Runs the client side of the simulation
    """
    implements(ISimulationClient)

    maxips = 100
    lasto = 0
    lastd = 0
    ecnt = 0
    dcnt = 0
    dtick = 0
    coiterating = 0
    pcid = 0
    enabledLoop = None
    disabledLoop = None
    enabledPeriod = 0.05
    disabledPeriod = 1.00

    def __init__(self):
        self.views = {}
        self.enabled = []
        self.disabled = []
        self.realms = {}
        self.areas = {}
        self.inview = {}
        self.areatranslations = {}
        self.arearotations = {}
        self.enabledState = {}
    
    
    def configure(self, nm, cf):
        """
        Configure the interface with this config file
        """


    def getItem(self, oid):
        return self.inview.get(oid)


    def addPlayerView(self, v):
        """ 
        Add v to the dict of player views 
        """
        self.views[v.from_connection] = v


    def getPlayerView(self, con):
        """ 
        Get the PlayerView from this connection if it is not
        available, then request it 
        """
        if self.views.has_key(con):
            return 
            
            
    def startService(self):
        """
        Start the cooperator!!
        """
        self.enabledLoop = LoopingCall(self.loopItems, True)
        self.disabledLoop = LoopingCall(self.loopItems, False)
        self.enabledLoop.start(self.enabledPeriod)
        self.disabledLoop.start(self.disabledPeriod)
        self.coiterating = 1
            
            
    def loopItems(self, enabled):
        """
        If start looping through enabled or disabled items depending
        on the value of enabled
        """
        # print "loopy", len(self.enabled), len(self.disabled)
        if enabled:
            return coiterate(ItemIterator(self, self.enabled))
        else:
            return coiterate(ItemIterator(self, self.disabled))


    def next(self):
        """
        iterate through the views
        """
        raise Exception("What?")
        try:
            start = self.lasto
            periter = max(len(self.enabled) / 5.0, 1)
            for x in range(periter):
                if len(self.enabled):
                    if self.lasto >= len(self.enabled):
                        self.lasto = 0
                    self.enabled[self.lasto].iterate(0.05)
                    if self.enabled[self.lasto].getID() == self.pcid:
                        for v in self.views.values():
                            v.updateCamera()
                    if not self.enabled[self.lasto].isEnabled():
                        # print "Disabling", self.enabled[self.lasto]
                        self.disabled.append(self.enabled[self.lasto])
                        self.enabled.remove(self.enabled[self.lasto])
                    self.lasto += 1
                    self.ecnt += 1
                    if self.lasto == start:
                        break
            if self.dtick >= 2 and len(self.disabled):
                self.dtick = 0
                if self.lastd >= len(self.disabled):
                    self.lastd = 0
                self.disabled[self.lastd].iterate(0.05)
                if self.lastd < len(self.disabled):
                    # ^^ yes we just checked that
                    # yes, during iterate on the last disabled
                    # item, it can change.
                    # really we should check if it changed states
                    # and then not continue
                    # really really we should do what the server does
                    if self.disabled[self.lastd].getID() == self.pcid:
                        for v in self.views.values():
                            v.updateCamera()
                    if self.disabled[self.lastd].isEnabled():
                        # print "Enabling", self.disabled[self.lastd]
                        self.enabled.append(self.disabled[self.lastd])
                        self.disabled.remove(self.disabled[self.lastd])
                    self.lastd += 1
                    self.dcnt += 1
            self.dtick += 1
            self.iteratePC(0.05)
        except:
            import traceback
            traceback.print_exc()
            # print "bad boy!"


    def enableItem(self, oid):
        """
        Enable an item
        """
        itm = self.inview[oid]
        if not itm in self.enabled:
            self.enabled.append(itm)
        if itm in self.disabled:
            self.disabled.remove(itm)


    def disableItem(self, oid):
        """
        Disable an item
        """
        itm = self.inview[oid]
        if not itm in self.disabled:
            self.disabled.append(itm)
        if itm in self.enabled:
            self.enabled.remove(itm)


    def iterate(self, time):
        try:
            # if not self.coiterating:
                # coiterate(self)
                # self.coiterating = 1
            for v in self.views.values():
                v.iterate(time)
            for r in self.realms.values():
                r.iterate(time)
            for a in self.areas.values():
                a.iterate(time)
            # update stats
            ps = self.parent.getLocalService(IPlayerClient)
            ps.updateStat(4,"IterEnable %d/%d"%(self.ecnt,len(self.enabled)))
            ps.updateStat(5,"IterDisable %d/%d"%(self.dcnt,len(self.disabled)))
            ps.updateStat(6,"Total %d"%len(self.inview))
            #self.client.UpdateStat(7,"Disabled %d"%len(self.disabled))
            self.dcnt = self.ecnt = 0
            if self.inview.has_key(self.pcid):
                p=self.inview[self.pcid].body.getPosition()
                a=self.inview[self.pcid].areas.keys()
            else:
                p=(0,0,0)
                a="None"
            ps.updateStat(8,"Pos: %.1f %.1f %.1f %s"%(p[0],p[1],p[2], a))
            return
        except:
            import traceback
            traceback.print_exc()
            # print "bad boy!"
      
      
    def runPhysics(self, time):
        for v in self.views.values():
            v.runPhysics(time)


    def iteratePC(self, time):
        for v in self.views.values():
            v.iteratePC(time)
