# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
The client side implimentation of the IGE Terrain editor plugin
"""

from twisted.internet import defer

from ogre.renderer import OGRE as ogre
from ogre.gui import CEGUI as cegui

from mv3d.phys.collider import castRay
from mv3d.util.classgen import ClassGenerator
from mv3d.client.ui.cegui import UserInterface
from mv3d.client.ui.ogre3d import OgreNode
from mv3d.client.ui.ogre3d import OgreObject
from mv3d.client.ui.image import Compositor


uiloaded = 0
moduiloaded = 0 


class TerrainProperties(UserInterface):
    """
    The client side interface to the IGE Terrain editor
    """
    def initialize(self, ui, i,ige):
        try:
            global uiloaded
            self.ui=ui
            self.interface=i
            self.ige=ige
            if not uiloaded:
                uiloaded=1
                w=self.ui.createComponent("WindowsLook/FrameWindow",
                    "TerrainProperties",[0.25,0.3],[0.5,0.5],[0.01,0.01],
                    [1.0,1.0],"Item Properties")
                iidl=ui.createComponent("WindowsLook/StaticText",
                    "TerrainProperties/IIDLText",[0.1,0.1],[0.3,0.1],[0.01,0.01],[1.0,1.0],'ID:')
                iid=ui.createComponent("WindowsLook/StaticText",
                    "TerrainProperties/IIDText",[0.4,0.1],[0.4,0.1],[0.01,0.01],[1.0,1.0],'')
                nml=ui.createComponent("WindowsLook/StaticText",
                    "TerrainProperties/NMLText",[0.1,0.2],[0.3,0.1],[0.01,0.01],[1.0,1.0],'Name:')
                nme=ui.createComponent("WindowsLook/Editbox",
                    "TerrainProperties/NMEEdit",[0.4,0.2],[0.4,0.1],[0.01,0.01],[1.0,1.0],'')
                modifyb=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/ModifyButton",[0.1,0.3],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Modify')
                paintb=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/PaintButton",[0.3,0.3],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Paint')
                self.texturesl=ui.createComponent("WindowsLook/MultiColumnList",
                    "TerrainProperties/TexturesList",[0.1,0.4],[0.5,0.4],[0.01,0.01],[1.0,1.0],'')
                addnum=ui.createComponent("WindowsLook/Editbox",
                    "TerrainProperties/TextureNum",[0.8,0.4],[0.2,0.1],[0.01,0.01],[1.0,1.0],'0')
                addtxt=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/AddTexture",[0.6,0.4],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Add')
                remtxt=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/RemTexture",[0.6,0.5],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Rem')
                uptxt=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/MoveUp",[0.6,0.6],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Move Up')
                downtxt=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/MoveDown",[0.6,0.7],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Move Down')
    
                cbtn=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/CButton",[0.4,0.9],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Cancel')
                abtn=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/AButton",[0.8,0.9],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Apply')
                okbtn=ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/OKButton",[0.6,0.9],[0.2,0.1],[0.01,0.01],[1.0,1.0],'Ok')
                self.texturesl.addColumn('Layer',0,cegui.UDim(0.2,0))
                self.texturesl.addColumn('Asset ID',1,cegui.UDim(0.8,0))
                w.addChildWindow(iidl)
                w.addChildWindow(iid)
                w.addChildWindow(nml)
                w.addChildWindow(nme)
                w.addChildWindow(modifyb)
                w.addChildWindow(paintb)
                w.addChildWindow(self.texturesl)
                w.addChildWindow(addtxt)
                w.addChildWindow(addnum)
                w.addChildWindow(remtxt)
                w.addChildWindow(uptxt)
                w.addChildWindow(downtxt)
                w.addChildWindow(cbtn)
                w.addChildWindow(abtn)
                w.addChildWindow(okbtn)
                
            self.window = self.ui.getComponent("TerrainProperties")
            self.deferred = 0
            self.modify = ModifyTerrain()
            self.texturesl = self.ui.getComponent("TerrainProperties/TexturesList")
            self.texturesl.resetList()
            self.texturesla = []
        except:
            import traceback
            traceback.print_exc()
            
            
    def start(self):
        self.ui.displayWindow(self.window)
        self.callbacks = []
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/ModifyButton","EventClicked",self,"Modify"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/PaintButton","EventClicked",self,"Paint"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/CButton","EventClicked",self,"Cancel"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/AButton","EventClicked",self,"Apply"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/OKButton","EventClicked",self,"Ok"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/AddTexture","EventClicked",self,"addTexture"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/RemTexture","EventClicked",self,"remTexture"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/MoveUp","EventClicked",self,"upTexture"))
        self.callbacks.append(self.ui.requestUICallback("TerrainProperties/MoveDown","EventClicked",self,"downTexture"))
        d=self.interface.callRemote("GetInfo")
        d.addCallback(self.GotInfo)
        return d


    def GotInfo(self, inf):
        self.ui.getComponent("TerrainProperties/IIDText").text=str(inf[0])
        self.ui.getComponent("TerrainProperties/NMEEdit").text=inf[1]
        self.item = self.ui.client.getInterface("Simulation Client").getItem(inf[0])
        self.textures = inf[2]
        self.updateTextureList()
        self.modify.initialize(self.ui, self.interface, self.ige, self.item)
        self.deferred = defer.Deferred()
        return self.deferred


    def stop(self):
        self.ui.hideWindow(self.window)
        for cb in self.callbacks:
            self.ui.stopUICallback(cb)
        
        
    def Show(self):
        """
        Show our window
        """
        self.ui.displayWindow(self.window)
        
        
    def Hide(self):
        """
        Hide our window
        """
        self.ui.hideWindow(self.window)
        
        
    def Cancel(self, args):
        try:
            self.stop()
            if self.deferred!=0:
                self.deferred.callback(0)
        except:
            import traceback
            traceback.print_exc()
            
            
    def Apply(self, args):
        try:
            nm=self.ui.getComponent("TerrainProperties/NMEEdit").text.c_str()
            self.interface.callRemote("setName",nm)
            self.sendTextureList()
        except:
            import traceback
            traceback.print_exc()


    def Ok(self, args):
        try:
            nm=self.ui.getComponent("TerrainProperties/NMEEdit").text.c_str()
            self.interface.callRemote("setName",nm)
            self.sendTextureList()
            self.stop()
            if self.deferred!=0:
                self.deferred.callback(1)
        except:
            import traceback
            traceback.print_exc()
            
            
    def Modify(self, args):
        try:
            self.modify.setMode("modify")
            d = self.modify.start()
            self.Hide()
            def done(r):
                self.Show()
            d.addCallback(done)
        except:
            import traceback
            traceback.print_exc()


    def Paint(self, args):
        try:
            self.modify.setMode("paint")
            d = self.modify.start()
            self.Hide()
            def done(r):
                self.Show()
            d.addCallback(done)
        except:
            import traceback
            traceback.print_exc()
            
            
    def updateTextureList(self):
        """
        Go through the list of textures we store and update 
        the texture listbox
        """
        self.texturesl.resetList()
        for n, t in self.textures.items():
            # print "Adding",n,t
            self.texturesla.append(self.ui.addRowToListBox(self.texturesl, [str(n), str(t)]))
            
            
    def getSelectedTexture(self):
        """
        Just returns the selected texture
        """
        return int(self.texturesl.getFirstSelectedItem().text.c_str())
        
        
    def remTexture(self, args):
        """
        Remove the selected texture
        """
        try:
            sel = self.getSelectedTexture()
            del self.textures[sel]
            self.updateTextureList()
        except:
            import traceback
            traceback.print_exc()
            
            
    def addTexture(self, args):
        """
        Add the selected texture
        """
        try:
            d= self.ige.getAssetSelector()
            def gotSelector(asel):
                asel.SetAssetTypes([ClassGenerator(modulename="mv3d.resource.asset",classname="ImageAsset")])
                d=asel.start()
                def gotSelection(asset):
                    num = int(self.ui.getComponent("TerrainProperties/TextureNum").text.c_str())
                    self.textures[num] = asset
                    self.updateTextureList()
                d.addCallback(gotSelection)
            d.addCallback(gotSelector)
        except:
            import traceback
            traceback.print_exc()
            
            
    def upTexture(self, args):
        """
        Move the selected texture up
        """
        try:
            sel = self.getSelectedTexture()
            if sel != 0:
                tmp = self.textures[sel - 1]
                self.textures[sel - 1] = self.textures[sel]
                self.textures[sel] = tmp
                self.updateTextureList()
        except:
            import traceback
            traceback.print_exc()
        
            
    def downTexture(self, args):
        """
        Move the selected texture down
        """
        try:
            sel = self.getSelectedTexture()
            if sel < len(self.textures) - 1:
                tmp = self.textures[sel + 1]
                self.textures[sel + 1] = self.textures[sel]
                self.textures[sel] = tmp
                self.updateTextureList()
        except:
            import traceback
            traceback.print_exc()


    def sendTextureList(self):
        """
        Send our texture list to the server
        """
        # print "sending Textures"
        return self.interface.callRemote("setTextures", self.textures)
        
        

class ModifyTerrain(UserInterface):
    """
    Define a UI to modify the height or textures on a terrain
    """
    def initialize(self, ui, i,ige, item):
        try:
            global moduiloaded
            self.ui = ui
            self.interface = i
            self.ige = ige
            self.item = item
            self.mode = "modify"
            if not moduiloaded:
                moduiloaded = 1
                # modify window
                w=self.ui.createComponent("WindowsLook/FrameWindow",
                    "TerrainProperties/ModifyWindow",[0.5,0.0],[0.5,0.08],[0.01,0.01], [1.0,1.0],"Move")
                modeb=self.ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/ModifyWindow/ModeButton",[0.0,0.25],[0.2,1.0],[0.01,0.01], [1.0,1.0],"Move")
                undob=self.ui.createComponent("WindowsLook/Button",
                    "TerrainProperties/ModifyWindow/UndoButton",[0.2,0.25],[0.1,1.0],[0.01,0.01], [1.0,1.0],"Undo")
                r1st=self.ui.createComponent("WindowsLook/StaticText",
                    "TerrainProperties/ModifyWindow/R1Text",[0.3,0.25],[0.05,1.0],[0.01,0.01], [1.0,1.0],"R1")
                r1t=self.ui.createComponent("WindowsLook/Editbox",
                    "TerrainProperties/ModifyWindow/R1Edit",[0.35,0.25],[0.20,1.0],[0.01,0.01], [1.0,1.0],"1.0")
                r2st=self.ui.createComponent("WindowsLook/StaticText",
                    "TerrainProperties/ModifyWindow/R2Text",[0.55,0.25],[0.05,1.0],[0.01,0.01], [1.0,1.0],"R2")
                r2t=self.ui.createComponent("WindowsLook/Editbox",
                    "TerrainProperties/ModifyWindow/R2Edit",[0.60,0.25],[0.20,1.0],[0.01,0.01], [1.0,1.0],"3.0")

                w.addChildWindow(modeb)
                w.addChildWindow(undob)
                w.addChildWindow(r1st)
                w.addChildWindow(r1t)
                w.addChildWindow(r2st)
                w.addChildWindow(r2t)
            self.window = self.ui.getComponent("TerrainProperties/ModifyWindow")
            self.deferred = 0
            self.mousedown = 0 
            self.adjpos = (5,5)
            self.undobuffers = 5
            self.undo = []
            self.submode = "Adjust"
            self.modifymodes = ["Adjust", "Flatten", "Smooth", "Roughen"]
            self.texturenames = {0:"Base", 1:"Texture 1",2:"Texture 2"}
            self.seltexture = 0
            self.comp = Compositor()
            #self.image = {}
            self.alpha = {}
            for n in self.texturenames.keys():
                if not self.item.alpha.has_key(n):
                    self.item.alpha[n] = self.comp.newImage((128,128))
                d = self.item.getVisualProperty("alphamaps").get(n, [])
                self.alpha[n] = self.comp.newImage((128,128), mode="L", data=d)
            self.modified = False
            self.visible = False
        except:
            import traceback
            traceback.print_exc()


    def start(self):
        self.ui.displayWindow(self.window)
        self.selectornode = OgreNode(self.ui.renderer)
        self.selectorobj = OgreObject(self.ui.renderer)
        class StupidMaterial:
            def getFilename(self):
                return "white"
        self.selectorobj.createManual("selector", verts=[(0,-1000,0), (0,1000,0)],
                    material=StupidMaterial(), renderoperation=ogre.RenderOperation.OT_LINE_LIST, recreate=True)
        self.selectornode.initialize("selectornode")
        self.selectornode.addObject(self.selectorobj)
        self.visible = True
        if self.mode == "modify":
            self.ccb=self.ui.requestUICallback("TerrainProperties/ModifyWindow/ModeButton","EventClicked",self,"modeClicked")
        else:
            self.ccb=self.ui.requestUICallback("TerrainProperties/ModifyWindow/ModeButton","EventClicked",self,"textureClicked")
        self.ccb=self.ui.requestUICallback("TerrainProperties/ModifyWindow/UndoButton","EventClicked",self,"undoClicked")
        self.ui.requestMouseCallback(self.mouseCB)
        self.deferred = defer.Deferred()
        return self.deferred


    def stop(self):
        self.ui.hideWindow(self.window)
        self.visible = False
        self.ui.stopMouseCallback(self.mouseCB)
        self.selectornode.finalize(1)
        self.selectorobj = None
        
        
    def setMode(self, mode):
        """
        Sets the mode for this modifier
        Valid modes are modify, which is for changing the shape of the terrain
        and paint, which is for painting textures on the terrain
        """
        self.mode = mode
        if mode == "modify":
            self.ui.getComponent("TerrainProperties/ModifyWindow/ModeButton").text = self.submode
        else:
            self.ui.getComponent("TerrainProperties/ModifyWindow/ModeButton").text = self.texturenames[self.seltexture]
        
        
    def modeClicked(self, args):
        """
        Toggle the current mode
        """
        try:
            mi = self.modifymodes.index(self.submode)
            mi +=1
            if mi == len(self.modifymodes):
                mi = 0
            self.submode = self.modifymodes[mi]
            self.ui.getComponent("TerrainProperties/ModifyWindow/ModeButton").text = self.submode
        except:
            import traceback
            print traceback.print_exc()

        
    def textureClicked(self, args):
        """
        Toggle the current texture
        """
        try:
            self.seltexture += 1
            if self.seltexture == len(self.texturenames):
                self.seltexture = 0
            self.ui.getComponent("TerrainProperties/ModifyWindow/ModeButton").text = self.texturenames[self.seltexture]
        except:
            import traceback
            print traceback.print_exc()

        
    def undoClicked(self, args):
        """
        Someone clicked the undo button. We need to go back if possible
        """
        try:
            if not len(self.undo):
                #no undos!
                return
            if self.mode == "modify":
                self.item.SetHeights(self.undo[-1])
                self.undo = self.undo[:-1]
        except:
            import traceback
            print traceback.print_exc()
        

    def mouseCB(self, btn, pos, rel, mods):
        try:
            if not self.visible:
                self.ui.stopMouseCallback(self.mouseCB)
                return
            if btn[0]!=0:
                if self.mousedown:
                    r1 = float(str(self.ui.getComponent("TerrainProperties/ModifyWindow/R1Edit").text.c_str()))
                    r2 = float(str(self.ui.getComponent("TerrainProperties/ModifyWindow/R2Edit").text.c_str()))
                    if self.mode == "modify":
                        if self.submode == "Adjust":
                            if self.adjpos!=-1:
                                self.adjustHeights(self.adjpos, r1, r2, -rel[1])
                        elif self.submode == "Flatten":
                            self.adjpos, p = self.getXZFromView(pos)
                            if self.adjpos != -1:
                                self.selectornode.setPosition(p)
                                self.flattenHeights(self.adjpos, r1, r2, self.targetheight)
                    else:
                        pos, p = self.getXZFromView(pos)
                        if pos != -1:
                            self.selectornode.setPosition(p)
                            self.drawTexture(pos, r1, r2)
                else:
                    self.mousedown = 1
                    self.adjpos, p = self.getXZFromView(pos)
                    if self.adjpos != -1:
                        self.selectornode.setPosition(p)
                    if self.mode == "modify":
                        if self.submode == "Flatten" and self.adjpos != -1:
                            self.targetheight = self.item.heights[int(self.adjpos[0])][int(self.adjpos[1])]
                        h = []
                        for r in self.item.heights:
                            h.append(list(r))
                        self.undo.append(h)
                        if len(self.undo) > self.undobuffers:
                            self.undo = self.undo[1:]
            else:
                if self.mousedown:
                    self.mousedown = 0
                    #send to server!
                    if self.mode == "modify":
                        self.interface.callRemote("setHeights", self.item.heights)
                    else:
                        if self.modified:
                            self.sendAlphasToServer()
                        self.modified = False
                crap, p = self.getXZFromView(pos)
                if crap != -1:
                    self.selectornode.setPosition(p)
            if btn[2]!=0:
                # cancel
                self.stop()
                self.deferred.callback(1)
        except:
            # print "ERRRROR"
            import traceback
            print traceback.print_exc()


    def adjustHeights(self, posv, radius1, radius2, amt):
        """
        Adjust the height of the terrain at pos by amt (positive = up).
        All height points will be adjusted if they are within the radius1
        If they are within radius2, they will be adjusted by a percentage
        based on how far they are away from radius1
        
        """
        #posv = (pos[0] * len(item.heights[0]), pos[1] * len(item.heights))
        radsq1 = pow(radius1, 2)
        radsq2 = pow(radius2, 2)
        for x in range(len(self.item.heights)):
            for z in range(len(self.item.heights[0])):
                d = pow(posv[0] - x, 2) + pow(posv[1] - z, 2)
                if d < radsq1:
                    self.item.heights[x][z] += amt
                elif d < radsq2:
                    ii = (radsq2 - radsq1)
                    if ii != 0:
                        falloff = float( (d - radsq1) / float(ii) )
                        if falloff != 0:
                            self.item.heights[x][z] += amt * (1.0 - falloff )
        self.item.destroyVisual()
        self.item.destroyPhysics()
        self.item.create()
        
        
    def flattenHeights(self, posv, radius1, radius2, height):
        """
        Adjust the height of the terrain at pos to height.
        All height points will be adjusted if they are within the radius1
        If they are within radius2, they will be adjusted by a percentage
        based on how far they are away from radius1
        
        """
        #posv = (pos[0] * len(item.heights[0]), pos[1] * len(item.heights))
        radsq1 = pow(radius1, 2)
        radsq2 = pow(radius2, 2)
        for x in range(len(self.item.heights)):
            for z in range(len(self.item.heights[0])):
                d = pow(posv[0] - x, 2) + pow(posv[1] - z, 2)
                if d < radsq1:
                    self.item.heights[x][z] = height
                elif d < radsq2:
                    ii = (radsq2 - radsq1)
                    if ii != 0:
                        falloff = float( (d - radsq1) / float(ii) )
                        if falloff != 0:
                            self.item.heights[x][z] = height * (1.0 - falloff ) + self.item.heights[x][z] * falloff
        self.item.destroyVisual()
        self.item.destroyPhysics()
        self.item.create()


    def getXZFromView(self, pos):
        """
        From an x & y coord on the screen (pos), return what height index (x,z) is at that pos
        Return -1 if no hit
        """
        ss=self.ui.renderer.getScreenDimensions()
        cam=self.ui.renderer.camera
        tx=pos[0] / float(ss[0])
        ty=pos[1] / float(ss[1])
        ray = cam.getCameraToViewportRay(tx , ty)
        pcid=self.ui.client.getInterface("Player Client").pcid
        area=self.ui.client.getInterface("Simulation Client").inview[pcid].areas.values()[0]
        spaces=[area.dspace,area.space]
        for space in spaces:
            r=castRay(space, ray.origin, ray.getPoint(20000), hitonly=[self.item.collider.geom])
            #print r
            if r.mingeom!=0:
                p = ray.getPoint(r.mindepth)
                pp = self.item.getXZAt(p)
                # print "hit at",p, pp
                return pp, p
        return -1, -1
        

    def drawTexture(self, pos, r1, r2):
        """
        Draw the selected texture at pos. Anything within r1 is drawn 100%
        anything within r2 is drawn with falloff.
        """
        self.modified = True
        sqr1 = pow(r1, 2)
        sqr2 = pow(r2, 2)
        tp = (pos[0] / float(len(self.item.heights)) * 128.0, pos[1] / float(len(self.item.heights[0])) * 128.0)
        for z in range(128):
            for x in range(128):
                d = pow(x - tp[0], 2) + pow(z - tp[1], 2)
                if d < sqr1:
                    for n, im in self.alpha.items():
                        if n == self.seltexture:
                            im.setPixel((x,z), 255)
                        else:
                            im.setPixel((x,z), 0)
                elif d < sqr2:
                    ii = (sqr2 - sqr1)
                    if ii != 0:
                        falloff = float( (d - sqr1) / float(ii) )
                        if falloff != 0:
                            for n, im in self.alpha.items():
                                v = im.getPixel((x,z))
                                if n == self.seltexture:
                                    im.setPixel((x,z), v * (falloff) + 255 * (1.0 - falloff))
                                else:
                                    im.setPixel((x,z), v * (falloff) + 0 * (1.0 - falloff))
        # now put the alpha in
        self.updateImages()
        
        
    def updateImages(self):
        """
        Update the alpha layer and convert to Ogre images, etc
        """
        for n, al in self.alpha.items():
            if n != 0:
                im = self.item.alpha[n]
                im.putAlpha(al)
                oim = im.convert("Ogre", name="Terrain %s alpha %d" % (self.item.getID(), n))
                self.comp.useInMaterial(oim, self.item.material.getMaterialName(), 0, n, 0)
            
            
    def sendAlphasToServer(self):
        """
        Convert all our alpha maps into lists and send them to the server
        """
        # return
        alphas = {}
        for n, al in self.alpha.items():
            if n != 0:
                alphas[n] = al.getData()
        self.interface.callRemote("setAlphaMaps", alphas)
