# Copyright (C) 2007-2008 Mortal Coil Games
# See LICENSE for details.
"""
classes for image compositor
"""


from ogre.renderer import OGRE as ogre


class Image:
    """
    class to hold a generic image
    assumed rgba 8 bits / channel
    """
    def __init__(self, **kw):
        self.Types = ["Ogre", "PIL", "CEGUI"]
        self.type = kw.get("type", "")
        self.image = kw.get("image", 0)
        if kw.get("file", 0) != 0:
            self.load(kw.get("file"))


    def getName(self):
        """
        If we are an ogre image, return name
        """
        if self.type == "Ogre":
            return self.image[0]
        return ""


    def load(self, fname):
        """
        Load the image from the file name 
        """
        if self.type == "PIL":
            from PIL import Image as image
            self.image = image.open(fname)
            return 1


    def setFrom(self, t, i):
        """
        You need to add a docstring here ! 
        """
        self.type = t
        self.image = i


    def getPixel(self, pos):
        """
        Takes pos (x,y) tuple
            returns a color (r,g,b,a) tuple
        """
        if self.image == 0:
            return 0
        if self.type == "PIL":
            return self.image.getpixel(pos)


    def setPixel(self, pos, col):
        """
        Takes pos (x,y) tuple
            and a color (r,g,b,a) tuple
            sets color at pos
        """
        if self.image == 0:
            return 0
        if self.type == "PIL":
            return self.image.putpixel(pos, col)


    def putAlpha(self, im):
        """
        Set our alpha channel to im!
        """
        if self.type == "PIL" and im.type == "PIL":
            self.image.putalpha(im.image)


    def getData(self):
        """
        Returns this image as a big list
        """
        if self.type == "PIL":
            return list(self.image.getdata())
        return []


    def convert(self, type, **args):
        """
        Convert the image from its current type
            to the new type..
        """
        if self.type == type:
            return self
        if self.type == "PIL":
            if type == "Ogre":
                import ctypes
                from ogre.renderer import OGRE as Ogre
                tm = Ogre.TextureManager.getSingleton()
                nam = args.get("name", "PILImage")
                grp = args.get("group", "General")
                if tm.resourceExists(nam):
                    tm.remove(nam)
                sz = self.image.size
                bsz = (sz[0] * sz[1])
                if self.image.mode == "RGB":
                    bsz *= 3
                else:
                    bsz *= 4
                sc = ctypes.c_char * bsz
                bm = sc(" ")
                bm.raw = self.image.tostring()
                ms = Ogre.MemoryDataStream(Ogre.CastVoidPtr(ctypes.addressof(bm)), bsz)
                pf = Ogre.PixelFormat.PF_A8B8G8R8
                if self.image.mode == "RGB":
                    pf = Ogre.PixelFormat.PF_B8G8R8
                tm.loadRawData(nam, grp, ms, sz[0], sz[1], pf, 
                    Ogre.TextureType.TEX_TYPE_2D, -1, 1.0)
                ni = newImage(type = "Ogre", image = [nam, grp])
                return ni
            if type == "CEGUI":
                import Ogre, ctypes, CEGUI
                tm = CEGUI.ImagesetManager.getSingleton()
                nam = args.get("name", "PILImage")
                grp = args.get("group", "General")
                ui = args.get("ui", 0)
                if tm.isImagesetPresent(nam):
                    tm.destroyImageset(nam)
                sz = self.image.size
                bsz = (sz[0] * sz[1])
                if self.image.mode == "RGB":
                    bsz *= 3
                else:
                    bsz *= 4
                sc = ctypes.c_char * bsz
                bm = sc(" ")
                bm.raw = self.image.tostring()
                bt = ui.guiRenderer.createTexture()
                #pf=bt.PF_RGBA#Ogre.PixelFormat.PF_A8B8G8R8
                #if self.image.mode=="RGB":
                    #pf=bt.PF_RGB#Ogre.PixelFormat.PF_B8G8R8
                #bt.loadFromMemory(Ogre.CastVoidPtr(ctypes.addressof(bm)),sz[0],sz[1],pf)#bt.PF_RGBA)
                    #bt.loadFromMemory(Ogre.CastVoidPtr(ctypes.addressof(bm)),sz[0],sz[1],bt.PF_RGBA)
                ms = Ogre.MemoryDataStream(Ogre.CastVoidPtr(ctypes.addressof(bm)), bsz)
                pf = Ogre.PixelFormat.PF_A8B8G8R8
                if self.image.mode == "RGB":
                    pf = Ogre.PixelFormat.PF_B8G8R8
                otm = Ogre.TextureManager.getSingleton()
                if otm.resourceExists(nam):
                    otm.remove(nam)
                otm.loadRawData(nam, grp, ms, sz[0], sz[1], pf, 
                    Ogre.TextureType.TEX_TYPE_2D, -1, 1.0)
                bt.loadFromFile(nam, grp)
                ims = tm.createImageset(nam, bt)
                ims.defineImage("full_image", CEGUI.Rect(0, 0, sz[0], sz[1]), CEGUI.Vector2(0, 0))
                return newImage(type = "CEGUI", image = [nam, bt])
        if self.type == "Ogre":
            if type == "PIL":
                import Ogre
                nam = self.image[0]
                grp = self.image[1]
                im = Ogre.TextureManager.getSingleton().getByName(nam)
                # print dir(im)
                qq = Ogre.Texture(im)
                if im:
                    from PIL import Image
                    sz = (im.getWidth(), im.getHeight())
                    b = im.getBuffer()
                    b.lock(b.HBL_READ_ONLY)
                    pb = b.getCurrentLock()
                    # print pb.data#[0:500]
                    im = Image.frombuffer("RGB", sz, pb.data)
                    b.unlock()
                    return newImage(type = "PIL", image = im)
                    #b.lock(b.HBU_WRITE_ONLY)
                    #pb=b.getCurrentLock()
                    #i=Image.open("screenshot_2006121607.png")
                    #pb.data#=i.tostring()
                    #image
                    #b.unlock()



def newImage(*a, **kw):
    """
    You need to add a docstring here ! 
    """
    return Image(*a, **kw)



class Compositor:
    """
    You need to add a docstring here ! 
    """
    def __init__(self):
        pass


    def newImage(self, size, type="PIL", mode="RGBA", data=[]):
        """
        Create a new image
        """
        import PIL.Image
        im = PIL.Image.new(mode, size)
        if len(data):
            im.putdata(data)
        return Image(type = "PIL", image = im)


    def blend(self, im1, im2, alpha):
        """
        return a new image that is im2 blended over im1 w/ constant alpha
        """
        import PIL.Image
        ni1 = im1.convert("PIL")
        ni2 = im2.convert("PIL")
        return Image(type = "PIL", image = PIL.Image.blend(ni1.image, ni2.image, alpha))


    def Composite(self, im1, im2, mask):
        """
        retuen a new image that is im2 blended over im1 with alpha specified
            by mask (which can be an image w/ alpha channel or 8bpp.
        """
        import PIL.Image
        ni1 = im1.convert("PIL")
        ni2 = im2.convert("PIL")
        nm = mask.convert("PIL")
        return Image(type = "PIL", image = PIL.Image.composite(ni1.image, ni2.image, nm))


    def createText(self, text, font="arial.ttf", fsize=15):
        """
        Create an image that consists of the text
        """
        import PIL.Image
        from PIL import ImageFont, ImageDraw
        font = ImageFont.truetype(font, fsize)
        size = font.getsize(text)
        image = PIL.Image.new("RGBA", size)
        draw = ImageDraw.Draw(image)
        draw.text((0, 0), text, font = font)
        return Image(type = "PIL", image = image)


    def useInMaterial(self, im, nm, tech=0, pss=0, tu=0):
        """
        Take im and use it in an Ogre material specified by nm.
            if specified, put it in the given technique/pass/textureunitstate
        """
        if not im.type == "Ogre":
            tnm = nm + "%d-%d-%d" % (tech, pss, tu)
            ni = im.convert("Ogre", name = nm)
        else:
            ni = im
        mat = ogre.MaterialManager.getSingleton().getByName(nm)
        if not mat.loaded:
            mat.load()
        mat.getTechnique(tech).getPass(pss).getTextureUnitState(tu).setTextureName(ni.image[0])


    def createNewMaterialFor(self, name, im, group=""):
        """
        Create a basic Ogre material for this image
        """
        if not im.type == "Ogre":
            return 0
        if group == "":
            group = ogre.ResourceGroupManager.DEFAULT_RESOURCE_GROUP_NAME
        m = ogre.MaterialManager.getSingleton().create(name, group)
        m.getTechnique(0).getPass(0).createTextureUnitState(im.getName())
        m.getTechnique(0).getPass(0).setSceneBlending(ogre.SBT_TRANSPARENT_ALPHA)
        return m
