# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
An object to represent a chunk of terrain
"""

import random

from array import array

from twisted.internet import defer
from twisted.internet.task import coiterate
from twisted.internet.threads import deferToThread

from ogre.renderer import OGRE as ogre
import ogre.addons.ogreforests as forests

from mv3d.util import math3d as VectorMath
from mv3d.net.pb import Cacheable
from mv3d.util.math3d import Vector
from mv3d.phys.collider import odeCollider, DeleteGeom
from mv3d.client.view.object import ObjectView
from mv3d.client.ui.ogre3d import OgreNode
from mv3d.client.ui.ogre3d import OgreObject
from mv3d.client.ui.image import Compositor


treecallback = None


class HeightGetterEntry:
    terrain = None
    def __init__(self, terrain, pos, size):
        self.terrain = terrain
        self.topleft = pos
        self.size = size
        self.bottomright = Vector(pos) + (size[0], 0, size[1])


    def isOnTerrain(self, pos):
        pp = Vector(pos) - self.topleft
        if pp[0] < 0 or pp[0] > self.size[0]:
            return False
        if pp[2] < 0 or pp[2] > self.size[1]:
            return False
        return True



class TerrainHeightGetter:
    def __init__(self):
        self.terrains = []


    def addTerrain(self, terrain, pos, size):
        self.terrains.append(HeightGetterEntry(terrain, pos, size))


    def getHeight(self, x, z):
        for t in self.terrains:
            if t.isOnTerrain((x,0,z)):
                return t.terrain.getTerrainHeight(x,z)
        # print "Height not found"
        return 0



#hack for pythonogre issue
theHeightGetter = TerrainHeightGetter()


class TreeAdder:
    offset = 0
    def __init__(self, terrain, count):
        # print "Created addder for", terrain, count
        self.terrain = terrain
        self.count = count
    
    def __iter__(self):
        return self
        

    def next(self):
        x=0
        z=0
        yaw = random.randrange(0, 360)

        for x in range(10):
            pp = (random.random(), random.random())
            offs = (self.terrain.getPosition()[0] + pp[0] * self.terrain.width, 
                self.terrain.getHeightAt(pp), self.terrain.getPosition()[2] + pp[1] * self.terrain.length)
            # x = random.randrange(self.terrain.getPosition()[0], 
                # self.terrain.getPosition()[0]+self.terrain.width)
            # x = random.randrange(0, self.width)
            # z = random.randrange(self.terrain.getPosition()[2], 
                # self.terrain.getPosition()[2]+self.terrain.length)
            # y = self.terrain.getTerrainHeight(x,z)
            # z = random.randrange(0, self.length)
            if offs[1]>0:
                break
        if offs[1]>0:
            # print "Placing tree at", offs, "for", list(self.terrain.getPosition()), self.terrain.width, self.terrain.length
            scale = random.randrange(9, 20) / 10
            
            t_type = random.randrange(0, len(self.terrain.treeModels))
            self.terrain.treeLoader.addTree(self.terrain.treeModels[t_type], 
                (offs[0], offs[1], offs[2]), ogre.Degree(yaw), scale)
        else:
            print "No trees for", self.terrain
        self.terrain.trees.update()
        self.offset += 1
        if self.offset == self.count:
            raise StopIteration



class Ogre3DTerrainView(ObjectView):
    """
    You need to add a docstring here ! 
    """
    def __init__(self):
        ObjectView.__init__(self)
        self.width = 0
        self.length = 0
        self.heights = []
        self.collider = 0
        self.destroyed = 0
        self.node = 0
        self.mesh = 0
        self.phyzlod = 1.0
        self.vislod = 1.0
        self.lastlodcheck = 0
        self.selent = 0
        self.grass = []
        self.tree = []
        self.grassobj = []
        self.grassnod = []
        self.treeobj = []
        self.treenod = []
        self.material = None


    def destroy(self):
        """
        You need to add a docstring here ! 
        """
        self.destroyVisual()
        if self.collider != 0:
            DeleteGeom(self.collider.geom)
            self.collider = 0
        self.destroyed = 1
        # ObjectView.__del__(self)


    def setupPhysics(self, a):
        """
        You need to add a docstring here ! 
        """
        if a == 0:
            return 0
        #print "Terrain is in",a.getID()
        if self.isEnabled():
            space = a.getSpace()
        else:
            space = a.getDisabledSpace()
        self.area = a
        if self.collider != 0:
            DeleteGeom(self.collider.geom)
            self.collider = 0
        if len(self.heights) == 0:
            return 0
        verts = []
        tris = []
        colsz = len(self.heights)
        for x in range(colsz):
            rowsz = len(self.heights[x])
            for z in range(rowsz):
                vx = (x / float(colsz - 0.99)) * self.width
                vy = self.heights[x][z]
                vz = (z / float(rowsz - 0.99)) * self.length
                verts.append((vx, vy, vz))
        tri = 0
        for x in range(colsz - 1):
            rowsz = len(self.heights[x])
            for z in range(rowsz - 1):
                tris.append((tri + rowsz + 1, tri + rowsz, tri))
                tris.append((tri + 1 + rowsz, tri, tri + 1))
                tri += 1
            tri += 1
        self.collider = odeCollider()
        self.collider.setType("Mesh")
        self.collider.setParameter("position", self.getPosition())
        #self.collider.setParameter("rotation",self.getRotation())
        self.collider.setParameter("vertexes", verts)
        self.collider.setParameter("triangles", tris)
        self.collider.setParameter("stationary", True)
        self.collider.setParameter("oid", self.getID())
        self.collider.Build(0, space)
        #print "Terrain",self.area.getID(),self.collider.geom
        return 1


    def destroyPhysics(self):
        """
        You need to add a docstring here ! 
        """
        self.collider.destroy()


    def create(self):
        """
        You need to add a docstring here ! 
        """
        # make a list of assets we need.
        al = []
        al.append(self.getVisualProperty("material"))
        if self.grass != 0:
            al.append(self.grass)
        for a in self.tree:
            al.append(a)
        for a in self.getVisualProperty("textures"):
            al.append(a)
        d = self.server.getInterface("Asset Client").AcquireAssets(al)
        if isinstance(d, defer.Deferred):
            d.addCallback(self.create_GotAssets)
            d.addErrback(self.server.Error)
            return d
        return self.create_GotAssets(d)


    def create_GotAssets(self, a): #grass, asset):
        """
        You need to add a docstring here ! 
        """
        material = a[self.getVisualProperty("material")]
        grass = a.get(self.grass, 0)
        tree = []
        for tr in self.tree:
            tree.append(a.get(tr, 0))
        assert material
        # make a copy of our material
        class FakeMaterialAsset:
            """
            You need to add a docstring here ! 
            """
            def __init__(self, name, mat):
                self.name = name
                self.mat = mat
            def getMaterialName(self):
                return self.name
        if self.material:
            #destroy it
            ogre.MaterialManager.getSingleton().remove(self.material.name)
        if not ogre.MaterialManager.getSingleton().getByName("Terrain %s Material" % str(self.getID())):
            self.material = FakeMaterialAsset("Terrain %s Material" % str(self.getID()), 
                ogre.MaterialManager.getSingleton().getByName(material.getMaterialName()).clone("Terrain %s Material" % str(self.getID())))
        else:
            self.material = FakeMaterialAsset("Terrain %s Material" % str(self.getID()), 
                ogre.MaterialManager.getSingleton().getByName("Terrain %s Material" % str(self.getID())))
        self.updateAlphas()
        self.updateTextures()
        #print "Createing Terrin"
        p = self.getPosition()
        i = self.server.GetRenderer()
        #msh=i.sceneManager.createManualObject(str(self.getID()))
        #msh.begin(asset.getFilename())
        #print colsz
        def genMesh():
            vertc = 0
            verts = []
            norms = []
            colors = []
            uv = []
            tris = []
            trinorms = []
            colsz = len(self.heights)
            for x in range(colsz):
                rowsz = len(self.heights[x])
                #print rowsz
                for z in range(rowsz):
                    vx = (x / float(colsz - 1)) * self.width
                    vy = self.heights[x][z]
                    vz = (z / float(rowsz - 1)) * self.length
                    #msh.position(vx,vy,vz)
                    verts.append((vx, vy, vz))
                    colors.append((1, 1, 1))
                    uv.append((x / float(colsz), z / float(rowsz)))
                    norms.append((0, 0, 0))
                    #msh.normal(0,1,0)
                    #msh.colour(1,1,1)
                    #msh.textureCoord(x/float(colsz),z/float(rowsz))
                    #verts.append(((x/float(colsz-0.98))*self.width,vy,(z/float(rowsz-0.98))*self.length))
                    vertc += 1
            tri = 0
            for x in range(colsz - 1):
                rowsz = len(self.heights[x])
                for z in range(rowsz - 1):
                    #msh.triangle(tri+rowsz+1, tri+rowsz, tri)
                    #msh.triangle(tri+1+rowsz, tri, tri+1)
                    tris.append((tri + rowsz + 1, tri + rowsz, tri))
                    v1 = verts[tri + rowsz + 1]
                    v2 = verts[tri + rowsz]
                    v3 = verts[tri]
                    s1 = VectorMath.subtractVectors(v1, v2)
                    s2 = VectorMath.subtractVectors(v2, v3)
                    trinorms.append(VectorMath.normalizeVector(VectorMath.CrossProduct(s1, s2)))
                    norms[tri + rowsz + 1] = VectorMath.addVectors(norms[tri + rowsz + 1], trinorms[ - 1])
                    norms[tri + rowsz] = VectorMath.addVectors(norms[tri + rowsz], trinorms[ - 1])
                    norms[tri] = VectorMath.addVectors(norms[tri], trinorms[ - 1])
                    tris.append((tri + 1 + rowsz, tri, tri + 1))
                    v1 = verts[tri + rowsz + 1]
                    v2 = verts[tri]
                    v3 = verts[tri + 1]
                    s3 = VectorMath.subtractVectors(v2, v1)
                    s4 = VectorMath.subtractVectors(v3, v2)
                    trinorms.append(VectorMath.normalizeVector(VectorMath.CrossProduct(s3, s4)))
                    norms[tri + rowsz + 1] = VectorMath.addVectors(norms[tri + rowsz + 1], trinorms[ - 1])
                    norms[tri] = VectorMath.addVectors(norms[tri], trinorms[ - 1])
                    norms[tri + 1] = VectorMath.addVectors(norms[tri + 1], trinorms[ - 1])
                    tri += 1
                tri += 1
            return (verts, norms, colors, uv, tris, trinorms)

        d = deferToThread(genMesh)
        def done(r):
            verts, norms, colors, uv, tris, trinorms = r
            self.verts = verts
            self.tris = tris
            i = self.server.GetRenderer()
            self.node = OgreNode(i, self.areas.values()[0].getNode().getNode())
            self.node.initialize(str(self.getID()))
            recreate = True
            if self.mesh == 0:
                self.mesh = OgreObject(i)
                # recreate = False
            self.mesh.createManual(str(self.getID()), verts = verts, tris = tris, norms = norms, 
                uvs = uv, material = self.material, recreate = recreate)#colors = colors, 
            self.mesh.castShadows(False)
            self.node.addObject(self.mesh)
            self.node.setPosition(self.getPosition())
            self.node.setRotation(self.getRotation())
            self.createTrees()
            for a in self.areas.values():
                self.server.CallLater(0.02, self.enterArea, a)
        d.addCallback(done)
        #print "Created a Terrain!"


    def destroyVisual(self):
        """
        You need to add a docstring here ! 
        """
        if self.node == 0:
            return
        #from pyogre import ogre
        #i=self.server.GetRenderer()
        #self.node.detachAllObjects()
        #i.sceneManager.rootSceneNode.removeChild("%sNode"%str(self.getID()))
        #i.sceneManager.destroySceneNode("%sNode"%str(self.getID()))
        #i.sceneManager.destroyManualObject(str(self.getID()))
        self.node.finalize(0)
        # self.mesh=0
        self.node = 0
        self.ent = 0
        #print "Destroyed a Terrain!"


    def getLOD(self, d):
        """
        Scale down the heightfield
            d=0.0-1.0 (0.0 means no mesh)
            There is really no problem I can see with going above 1,0, and
            eventually, this function may want to implement some sort of
            blerping or splining.
        """
        if d == 0:
            return []
        if d == 1:
            return self.heights
        heights = []
        for x in range(int(len(self.heignts) * d)):
            xo = int(x / d)
            row = []
            for z in range(int(len(self.heights[xo]) * d)):
                zo = int(z / d)
                row.append(self.heights[xo][zo])
            heights.append(row)
        return heights


    def enterArea(self, a):
        """
        You need to add a docstring here ! 
        """
        self.setupPhysics(a)
        return 1


    def leaveArea(self, a):
        """
        You need to add a docstring here ! 
        """
        self.destroyVisual()
        self.destroyPhysics()


    def getWidth(self):
        """
        You need to add a docstring here ! 
        """
        return self.width


    def setWidth(self, w):
        """
        You need to add a docstring here ! 
        """
        self.width = w
        self.updateClients("setWidth", w)
        self.getView().setWidth(w)
        self.setupPhysics(self.area)


    def observe_setWidth(self, w):
        """
        You need to add a docstring here ! 
        """
        self.setWidth(w)


    def getLength(self):
        """
        You need to add a docstring here ! 
        """
        return self.length


    def setLength(self, l):
        """
        You need to add a docstring here ! 
        """
        self.length = l
        self.getView().setLength(l)
        self.setupPhysics(self.area)


    def observe_setLength(self, l):
        """
        You need to add a docstring here ! 
        """
        self.setLength(l)


    def getHeight(self, p):
        """
        You need to add a docstring here ! 
        """
        return self.heights[p[0]][p[1]]


    def setHeight(self, p, h):
        """
        You need to add a docstring here ! 
        """
        self.heights[p[0]][p[1]] = h
        self.getView().setHeight(p, h)
        self.setupPhysics(self.area)


    def observe_setHeight(self, p, h):
        """
        You need to add a docstring here ! 
        """
        self.setHeight(p, h)


    def getHeights(self):
        """
        You need to add a docstring here ! 
        """
        return self.heights


    def getHeightAt(self, p):
        """ 
        Gets the height at p (x,z) tuple 
        interpolate.. 0<x<1 0<y<1
        """
        th = len(self.heights) - 1
        tw = len(self.heights[0]) - 1
        ww = p[0] * tw
        hh = p[1] * th
        iw = int(ww)
        ih = int(hh)
        pw = ww - iw
        ph = hh - ih
        def interp(a, b, p):
            """
            You need to add a docstring here ! 
            """
            return (a * (1.0 - p)) + (b * p)
        if ih >= th:
            if iw >= tw:
                return self.heights[tw][th]
            return interp(self.heights[iw][th], self.heights[iw + 1][th], pw)
        if iw >= tw:
            return interp(self.heights[tw][ih], self.heights[tw][ih + 1], ph)
        h00 = self.heights[iw][ih]
        h01 = self.heights[iw][ih + 1]
        h10 = self.heights[iw + 1][ih]
        h11 = self.heights[iw + 1][ih + 1]
        ww1 = interp(h00, h01, ph)
        ww2 = interp(h10, h11, ph)
        return interp(ww1, ww2, pw)
        
        
    def getTerrainHeight(self, x, z):
        # print "getTerrainHeight", x, z, self.getPosition()[0], self.getPosition()[2], self.width, self.length
        p = Vector((x,0,z)) - self.getPosition()
        xx = p[0] / float(self.width)
        zz = p[2] / float(self.length)
        # if xx>1.0 or xx<0.0 or zz<0.0 or zz>1.0:
            # print xx, zz, p[0], p[1], x, z, list(self.getPosition())
        return self.getHeightAt((xx,zz))


    def getXZAt(self, p):
        """
        Gets the height array index x, z at real world pos p. Returns -1 if out of bounds
        """
        # print p[0], p[2], self.getPosition()[0], self.getPosition()[2]
        lp = VectorMath.subtractVectors(p, self.getPosition())
        # print lp[0]/self.width, lp[2]
        if lp[0] < 0 or lp[0] > self.width:
            # print "wid",lp[0], self.width
            return (0,0)
        if lp[2] < 0 or lp[2] > self.length:
            # print "len", lp[2], self.length
            return (0,0)
        ix = int(lp[0] / float(self.width)) * float(len(self.heights))
        iz = int(lp[2] / float(self.length)) * float(len(self.heights[0]))
        return (ix, iz)


    def setHeights(self, h):
        """
        You need to add a docstring here ! 
        """
        self.heights = h
        # self.getView().setHeights(h)
        # self.setupPhysics(self.area)
        self.destroyVisual()
        self.destroyPhysics()
        self.create()


    def observe_setHeights(self, h):
        """
        You need to add a docstring here ! 
        """
        self.setHeights(h)


    def updateHeights(self, p1, p2, h):
        """
        You need to add a docstring here ! 
        """
        ww = p2[0] - p1[0]
        ll = p2[1] - p1[1]
        for x in range(ww):
            for z in range(ll):
                self.heights[p1[0] + x][p1[1] + z] = h[x][z]
        self.setupPhysics(self.area)


    def isEnabled(self):
        """
        You need to add a docstring here ! 
        """
        return 0


    def iterate(self, t):
        """
        You need to add a docstring here ! 
        """
        if hasattr(self, "trees"):
            self.trees.update()
            self.grass.update()
            self.grassLoader.updateAnimation()   


    def observe_updateHeights(self, p1, p2, h):
        """
        You need to add a docstring here ! 
        """
        self.updateHeights(p1, p2, h)


    def select(self):
        """
        You need to add a docstring here ! 
        """
        #print "Select",self.getID()
        if self.node == 0:
            return
        if self.selent != 0:
            return
        self.selent = OgreObject(self.server.GetRenderer())
        # self.selent.createBillboardSet("Select_%s"%str(self.getID()),material="SelectedMaterial",
            # billboards=[{"position":(0,40,0)}])
        self.selent.createWireBoundingBox("Select_%s" % str(self.getID()), obj = self.mesh, material = "Examples/Hilite/Yellow")
        self.node.addObject(self.selent)


    def deselect(self):
        """
        You need to add a docstring here ! 
        """
        #print "Deselect",self.getID()
        if self.selent != 0:
            self.selent.destroy()
        self.selent = 0


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        h = array("f", state["heights"])
        dw = state["datawidth"]
        state["heights"] = [h[x:x+dw] for x in range(0,len(h),dw)]
        Cacheable.setCopyableState(self, state)
        #self.server.CallLater(0.02,self.SetupVisual)
        #self.create()
        #print "Got a terrain!"


    def updateAlphas(self):
        """
        Make all the alpha maps in our material proper
        """
        c = Compositor()
        self.alpha = {}
        for n, map in self.getVisualProperty("alphamaps").items():
            im = c.newImage((128, 128), "PIL", "L", map)
            self.alpha[n] = c.newImage((128, 128), "PIL", "RGBA")
            self.alpha[n].putAlpha(im)
            nim = self.alpha[n].convert("Ogre", name = "Terrain %s alpha %d" % (self.getID(), n))
            c.useInMaterial(nim, self.material.name, 0, n, 0)


    def updateTextures(self):
        """
        Update the texture maps on our material. Grab the assets if needed, 
        so this may return a deferred
        """
        textids = self.getVisualProperty("textures")
        d = self.server.getInterface("Asset Client").AcquireAssets(textids.values())
        def gotAssets(r, textids):
            textures = {}
            for n, txt in textids.items():
                #print "applying",r[txt].getImageName(), n, txt, self.material.mat.getTechnique(0).getPass(n).getNumTextureUnitStates()
                if n == 0:
                    st = 0
                else:
                    st = 1
                self.material.mat.getTechnique(0).getPass(n).getTextureUnitState(st).setTextureName(r[txt].getImageName())


        if isinstance(d, defer.Deferred):
            d.addCallback(gotAssets, textids)
            return d
        return gotAssets(d, textids)


    def observe_setVisualProperty(self, p, v):
        """
        Set a visual property and  update if needed
        """
        ObjectView.observe_setVisualProperty(self, p, v)
        if p == "alphamaps":
            self.updateAlphas()
        if p == "textures":
            self.updateTextures()
            
            
    def createTrees(self):
        global treecallback
        # print "+++++++++++ Create trees", self
        renderer = self.server.GetRenderer()
        self.trees = forests.PagedGeometry()
        self.trees.setCamera( renderer.camera )
        # self.trees.setPageSize( 500 )
        self.trees.setPageSize( 100 )
        self.trees.setBounds( ogre.FloatRect(self.getPosition()[0], self.getPosition()[2], 
            self.getPosition()[0]+self.width, self.getPosition()[2]+self.length) )
        
        if not ogre.TextureManager.getSingleton().resourceExists("DynamicTexture"):
            self.mTex = ogre.TextureManager.getSingleton().createManual(
                            "DynamicTexture",
                            ogre.ResourceGroupManager.DEFAULT_RESOURCE_GROUP_NAME,
                            ogre.TEX_TYPE_2D,
                            16, 16,
                            0,
                            ogre.PF_BYTE_BGRA,
                            ogre.TU_DYNAMIC)
    
        ##Set up LODs
        ##self.trees.addDetailLevelEntityPage(50, 50)
        self.trees.addDetailLevelBatchPage(60, 10)
        self.trees.addDetailLevelImpostorPage(2000, 50)
        ##Set up a TreeLoader for easy use
        self.treeLoader = forests.TreeLoader3D( self.trees, ogre.FloatRect(self.getPosition()[0], self.getPosition()[2], 
            self.width, self.length) )
        self.trees.setPageLoader(self.treeLoader)
        # self.treeLoader.setHeightFunction(self, "getTerrainHeight")
        
    
        ##And add 20,000 trees to the scene through the "easy-to-use" TreeLoader class
        if not renderer.sceneManager.hasEntity("MyTree"):
            palm1 = renderer.sceneManager.createEntity("MyTree", "KinkyPalm.mesh")
            palm2 = renderer.sceneManager.createEntity("pineTree", "StraightPalm.mesh")
            palm3 = renderer.sceneManager.createEntity("palmBendy", "BendyPalm.mesh")
            bush1 = renderer.sceneManager.createEntity("bush1", "BigDarkBush.mesh")
            birch = renderer.sceneManager.createEntity("birch", "birch.mesh")
            benTree = renderer.sceneManager.createEntity("benTree", "benTree.mesh")
            spot_shroom = renderer.sceneManager.createEntity("redShroom", "spotty_shroom.mesh")
            blue_shroom = renderer.sceneManager.createEntity("blueShroom", "blue_shroom.mesh")
            beech = renderer.sceneManager.createEntity("beech", "beech_tree.mesh")
            newTree = renderer.sceneManager.createEntity("newTree", "Pine.mesh")
        else:
            palm1 = renderer.sceneManager.getEntity("MyTree")
            palm2 = renderer.sceneManager.getEntity("pineTree")
            palm3 = renderer.sceneManager.getEntity("palmBendy")
            bush1 = renderer.sceneManager.getEntity("bush1")
            birch = renderer.sceneManager.getEntity("birch")
            benTree = renderer.sceneManager.getEntity("benTree")
            spot_shroom = renderer.sceneManager.getEntity("redShroom")
            blue_shroom = renderer.sceneManager.getEntity("blueShroom")
            beech = renderer.sceneManager.getEntity("beech")
            newTree = renderer.sceneManager.getEntity("newTree")
        self.treeModels = [ bush1, newTree, birch, benTree, beech ]
        # fix the damn rotations

        #self.treeLoader.setColorMap(self.mTex)

        ##Grass
        
        # trees = new PagedGeometry(camera, 50, ogre.FloatRect(0, 0, WORLD_SIZE, WORLD_SIZE))
        def later(ignored):
            return coiterate(TreeAdder(self, 100))
        if not treecallback:
            treecallback = later(1)
        else:
            treecallback.addCallback(later)
        ##initialize the PagedGeometry engine
        self.grass = forests.PagedGeometry(renderer.camera, 50)
        self.grass.setBounds(ogre.FloatRect(self.getPosition()[0], self.getPosition()[2], 
                self.getPosition()[0]+self.width, self.getPosition()[2]+self.length))
        ##Set up LODs
        self.grass.addDetailLevelGrassPage(200.0,0.0)
        
        ##Set up a TreeLoader for easy use
        self.grassLoader = forests.GrassLoader(self.grass)
        self.grass.setPageLoader(self.grassLoader)
        self.grassLoader.setHeightFunction(theHeightGetter, "getHeight")
        theHeightGetter.addTerrain(self, self.getPosition(), (self.width, self.length))

        ## iterate grass materials
        for nam in ["grass", "dry_grass", "tall_grass"]:
            grassLayer = self.grassLoader.addLayer(nam)
            
            grassLayer.setAnimationEnabled(True)
            grassLayer.setSwaySpeed(1.5)
            grassLayer.setSwayLength(0.08)
            grassLayer.setSwayDistribution(10.0)
            
            grassLayer.setDensity(0.01)
            
            grassLayer.setMapBounds(ogre.FloatRect(self.getPosition()[0], self.getPosition()[2], 
                self.getPosition()[0]+self.width, self.getPosition()[2]+self.length))
            
            grassLayer.setColorMap("terrain_grass.jpg")
            # grassLayer.setColorMap(self.mTex)
            grassLayer.setColorMapFilter(forests.MAPFILTER_BILINEAR)
            
            # grassLayer.setDensityMap("densitymap.png")
            # grassLayer.setDensityMapFilter(forests.MAPFILTER_NONE)
        
            ##grassLayer.setMinimumSize(0.5,0.5)
            ##grassLayer.setMaximumSize(1.0, 1.0) 
            
            grassLayer.setMinimumSize(4.0,2.0)
            grassLayer.setMaximumSize(8.0,5.0)
            
            grassLayer.setFadeTechnique(forests.FADETECH_ALPHA)
