# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
""" 
    This represents what a client can view of an object 
"""

import ode

from ogre.renderer import OGRE as ogre

from mv3d.phys.collider import odeCollider
from mv3d.net.security import Securable
from mv3d.net.pb import Cacheable
from mv3d.util.classgen import ClassGenerator


class ObjectView(Securable, Cacheable):
    def __init__(self):
        self.visualproperties={}
        self.pos=(0,0,0)
        self.rot=(1,0,0,0)
        self.lvel=(0,0,0)
        self.avel=(0,0,0)
        self.vpo=(0,0,0)
        self.areas={}
        self.realms={}
        self.iid=-1
    
    
    def getID(self):
        return self.iid
  
  
    def setID(self,i):
        self.iid=i
        self.updateClients("setID",i)
    
    
    def observe_setID(self, i):
        self.iid=i


    def setVisualProperty(self, p, v):
        self.visualproperties[p]=v
        self.updateClients("setVisualProperty",p,v)


    def getVisualProperty(self, p,dv=None):
        return self.visualproperties.get(p,dv)


    def observe_setVisualProperty(self, p, v):
        self.visualproperties[p]=v


    def getPosition(self):
        return self.pos


    def setPosition(self, p):
        self.pos=p
        self.updateClients("setPosition",p)


    def observe_setPosition(self, p):
        self.pos=p


    def getRotation(self):
        return self.rot


    def setRotation(self, r):
        self.rot=r
        self.updateClients("setRotation",r)


    def observe_setRotation(self, r):
        self.rot=r


    def getLinearVelocity(self):
        return self.levl


    def setLinearVelocity(self, lv):
        self.lvel=lv
        self.updateClients("setLinearVelocity",lv)


    def observe_setLinearVelocity(self, lv):
        self.lvel=lv


    def getAngularVelocity(self, av):
        return self.avel


    def setAngularVelocity(self, av):
        self.avel=av
        self.updateClients("setAngularVelocity",av)


    def observe_setAngularVelocity(self, av):
        self.avel=av


    def updatePhysics(self, p,r,lv,av):
        self.pos=p
        self.rot=r
        self.lvel=lv
        self.avel=av
        self.updateClients("updatePhysics",p,r,lv,av)


    def observe_updatePhysics(self, p,r,lv,av):
        self.pos=p
        self.rot=r
        self.lvel=lv
        self.avel=av


    def getViewPointOffset(self):
        return self.vpo


    def setViewPointOffset(self, vpo):
        self.vpo=vpo
        self.updateClients("setViewPointOffset",vpo)


    def observe_setViewPointOffset(self, vpo):
        self.vpo=vpo


    def getAreas(self):
        return self.areas


    def observe_addArea(self, a):
        self.areas[a.getID()]=a
        # self.enterArea(a) #??


    def observe_remArea(self, a):
        del self.areas[a]
        self.leaveArea(a)


    def getRealms(self):
        return self.realms


    def addRealm(self, a):
        self.realms[a.getID()]=a
        self.updateClients("addRealm",a)


    def observe_addRealm(self, a):
        self.realms[a.getID()]=a


    def remRealm(self, a):
        if isinstance(a,tuple):
            if self.realms.has_key(a):
                del self.realms[a]
                self.updateClients("remRealm",a)
                return 1
            return 0
        if self.realms.has_key(a.getID()):
            del self.realms[a.getID()]
            self.updateClients("remRealm",a.getID())
            return 1
        return 0


    def observe_remRealm(self, a):
        del self.realms[a]
    

    def iterate(self, i):
        pass
    
    
    def isEnabled(self):
        return 1


    def create(self):
        pass


    def destroy(self):
        pass


    def getStateToCacheAndObserveFor(self, perspective, observer):
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        Cacheable.setCopyableState(self, state)


    def getClassGenerator(self):
        return ClassGenerator(classname="ObjectView",modulename="mv3d.client.view.ObjectView",
            sourceclass="mv3d.server.model.object.ObjectView")



created = 0



class Ogre3DGroundPlaneObjectView(ObjectView):
    """
    You need to add a docstring here ! 
    """
    def __init__(self):
        ObjectView.__init__(self)
        self.node = 0


    def __del__(self):
        self.destroyMesh()
        self.destroyed = 1
        Cacheable.__del__(self)


    def enterArea(self, a):
        """
        You need to add a docstring here ! 
        """
        #rid = a.getID()[0]
        r = a.realm
        if not r:
            #print "not R"
            return 0
        w = r.world
        #self.body = ode.Body(w)
        self.setupBody(w, a)
        return 1


    def setupBody(self, w, a):
        """
        You need to add a docstring here ! 
        """
        space = a.getDisabledSpace()
        self.cs = odeColliderSet()
        c = odeCollider()
        c.setType("Plane")
        #c.setParameter("size", (2000, 50, 2000))
        #c.setParameter("position", (0,  - 25, 0))
        c.setParameter("stationary", True)
        self.cs.AddCollider(c)
        self.body = ode.Body(w)
        self.body.setPosition(self.getPosition())
        self.body.setQuaternion(self.getRotation())
        self.body.setGravityMode(False)
        self.cs.Build(self.body, space)


    def create(self):
        """
        You need to add a docstring here ! 
        """
        global created
        p = self.getPosition()
        plane = ogre.Plane()
        plane.normal = ogre.Vector3.UNIT_Y
        plane.d = -2
        if not created:
            if self.server.GetRenderer().renderWindow != 0:
                mm = ogre.MeshManager.getSingleton()
                mm.createPlane('MyPlane', 'General', plane, 15000.0, 15000.0, 
                    20, 20, True, 1, 600, 600, (0, 0, 1))
            created = 1
        i = self.server.GetRenderer()
        self.entity = i.createEntity(str(self.getID()), 'MyPlane')
        if not isinstance(self.entity, int):
            self.entity.setMaterialName(self.getVisualProperty("material"))
            self.entity.castShadows = False
        if i.sceneManager != 0:
            self.node = i.sceneManager.rootSceneNode.createChildSceneNode("%sNode" % str(self.getID()))
            if self.entity:
                self.node.attachObject(self.entity)
            self.node.position = p #0.0, 99.0, 0.0
        for a in self.areas.values():
            self.enterArea(a)


    def destroyMesh(self):
        """
        You need to add a docstring here ! 
        """
        if self.node ==0: 
            return
        i = self.server.GetRenderer()
        self.node.detachAllObjects()
        i.sceneManager.rootSceneNode.removeChild("%sNode" % str(self.getID()))
        i.sceneManager.destroySceneNode("%sNode" % str(self.getID()))
        if self.entity:
            i.sceneManager.destroyEntity(str(self.getID()))
        self.node = 0
        self.ent = 0
        #print "Destroying", self.getID(), self


    def isEnabled(self):
        """
        You need to add a docstring here ! 
        """
        return 0


    def observe_setPosition(self, p):
        """
        You need to add a docstring here ! 
        """
        ObjectView.observe_setPosition(self, p)
        if self.node != 0:
            self.node.position = p


    def observe_setRotation(self, r):
        """
        You need to add a docstring here ! 
        """
        ObjectView.observe_setRotation(self, r)
        if self.node != 0:
            self.node.orientation = ogre.Quaternion(r[0], r[1], r[2], r[3])


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        Cacheable.setCopyableState(self, state)
        #self.create()


    def stoppedObserving(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        Cacheable.stoppedObserving(self, perspective, observer)

