# Copyright (C) 2006-2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

import os
import ctypes

from urlparse import urlparse

from twisted.internet import defer

from mv3d.util.datastore.store import attributes, text, classHash, autoStore
from mv3d.resource.asset import MaterialAsset, MeshAsset
from mv3d.net.pb import Cacheable, withClientUpdate
from mv3d.resource.url import URLMeshAsset, URLMaterialAsset
from mv3d.util.classgen import ClassGenerator
from mv3d.util.conductor import IPlayerClient


class LocalOgreMaterialAsset(MaterialAsset):
    """
    An asset that is a local file and is an ogre material 
    """
    datastoreAttributes = attributes(MaterialAsset, filename=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, filename=None, server=None, datastore=None):
        MaterialAsset.__init__(self)
        self.filename = filename
        self.datastore = datastore
        self.server = server


    def getFilename(self):
        """
        You need to add a docstring here ! 
        """
        return self.filename


    @autoStore
    @withClientUpdate
    def setFilename(self, f):
        """
        You need to add a docstring here ! 
        """
        self.filename = f


    observe_setFilename = setFilename


    def getMaterialName(self):
        """
        You need to add a docstring here ! 
        """
        return self.filename


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        #print "------ ASTATE",perspective, observer
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.setCopyableState(self, state)



def newURLOgreMeshAsset(assetgroup, url, localfile="", name="", depends=[]):
    """
    You need to add a docstring here ! 
    """
    d = assetgroup.newAsset(ClassGenerator(classname = "URLOgreObjectAsset", 
        modulename = "mv3d.resource.ogre3d"))

    for dd in depends:
        assert not isinstance(dd, defer.Deferred)

    def finishUp(tm, u, localfile, name, depends):
        """
        You need to add a docstring here ! 
        """
        tm.setURL(url)
        if localfile != "":
            tm.setLocalFile(localfile)
        if name != "":
            tm.setName(name)
        for d in depends:
            if isinstance(d, tuple):
                tm.addDependency(d)
            else:
                tm.addDependency(d.getID())
        return tm
    d.addCallback(finishUp, url, localfile, name, depends)
    return d



class URLOgreObjectAsset(URLMeshAsset):
    """
    You need to add a docstring here ! 
    """
    datastoreAttributes = attributes(URLMeshAsset, meshname=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1

    materialname = ""
    readied = None


    def __init__(self, datastore=None, url=None, localfile=None, basedir=u"Extern",
            server=None, meshname=None):
        URLMeshAsset.__init__(self, datastore, url, localfile, basedir)
        self.meshname = meshname
        server = self.server


    def setURL(self, u):
        """
        You need to add a docstring here ! 
        """
        p = urlparse(u)
        path = p[2]
        pp = path.split("/")
        self.localfile = pp[ - 1]
        self.meshname = pp[ - 1]
        URLMeshAsset.setURL(self, u)


    def setFilename(self, f):
        """
        You need to add a docstring here ! 
        """
        self.meshname = os.path.split(f)[1]
        URLMeshAsset.setFilename(self, f)


    observe_setURL = setURL
    observe_setFilename = setFilename


    def create(self, name):
        """
        You need to add a docstring here ! 
        """
        from ogre.renderer import OGRE as Ogre
        #print "Calling create for",self.meshname
        assert self.parent, "got no parent!"
        i = self.parent.parent.getLocalService(IPlayerClient).getRenderer()
        if not Ogre.ResourceGroupManager.getSingleton().resourceExists("General", self.meshname):
            # register the resource group
            Ogre.ResourceGroupManager.getSingleton().addResourceLocation(self.getFullPath(), "FileSystem", 
                "General")
        assert Ogre.ResourceGroupManager.getSingleton().resourceExists(
            "General", self.meshname), "Couldn't find resource!"
        return i.createEntity(str(name), self.meshname)


    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState



def newURLOgreMaterialAsset(assetgroup, url, materialname, localfile="", name="", depends=[]):
    tm = assetgroup.newAsset(ClassGenerator(classname="URLOgreMaterialAsset",
        modulename="mv3d.resource.ogre3d"))

    def finishUp(tm, u, m, l, n, dp):
        tm.setURL(url)
        tm.setMaterialName(materialname)
        if localfile != "":
            tm.setLocalFile(localfile)
        if name != "":
            tm.setName(name)
        else:
            tm.setName(materialname)
        for d in depends:
            if isinstance(d, tuple):
                tm.addDependency(d)
            else:
                tm.addDependency(d.getID())
        return tm
    if isinstance(tm, defer.Deferred):
        tm.addCallback(finishUp, url, materialname, localfile, name, depends)
        return tm
    return finishUp(tm, url, materialname, localfile, name, depends)



class URLOgreMaterialAsset(URLMaterialAsset):
    """
    An asset that is downloadable by URL and represents
    a material script
    """
    datastoreAttributes = attributes(URLMaterialAsset, materialname=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1

    materialname = ""
    readied = False


    def __init__(self, datastore=None, url=None, localfile=None, basedir=u"Extern",
            server=None, materialname=None):
        URLMaterialAsset.__init__(self, datastore, url, localfile, basedir)
        self.materialname = materialname
        self.server = server


    def setURL(self, u):
        URLMaterialAsset.setURL(self, u)
        p=urlparse(u)
        path=p[2]
        pp=path.split("/")
        self.localfile=pp[-1]


    def getMaterialName(self):
        return self.materialname


    @autoStore
    @withClientUpdate
    def setMaterialName(self, m):
        self.materialname=m


    observe_setMaterialName = setMaterialName


    def acquireAsset(self, **kw):
        d = URLMaterialAsset.acquireAsset(self, **kw)
        def done(r):
            self.readyAsset()
            return r
        d.addCallback(done)
        return d


    def urlHaveAsset(self):
        if not URLMaterialAsset.urlHaveAsset(self):
            return 0
        self.readyAsset()
        return 1


    def readyAsset(self):
        # print "THINKING ABOUT READYING"
        if not self.readied:
            # print "Registering materials from", self.getFullFile()
            self.readied = 1
            from ogre.renderer import OGRE as Ogre
            f = open(self.getFullFile())
            s = f.read()
            f.close()
            # print "data:", s[:100]
            sc = ctypes.c_char * len(s)
            cp = sc(" ")
            for x in range(len(s)):
                cp[x] = s[x]
            ds=Ogre.MemoryDataStream(Ogre.CastVoidPtr(ctypes.addressof(cp)), len(s))
            Ogre.MaterialManager.getSingleton().parseScript(ds, "General")


    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState



class LocalOgreObjectAsset(MeshAsset):
    """
    An asset that represents an ogre object
    """
    datastoreAttributes = attributes(MeshAsset, filename=text, type=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1

    Types = ("Mesh", "Light", "Particles")

    def __init__(self, filename=None, datastore=None, server=None, ype=None):
        MeshAsset.__init__(self)
        self.filename = filename
        self.type = type
        self.datastore = datastore
        self.server = server


    def getFilename(self):
        """
        You need to add a docstring here ! 
        """
        return self.filename


    @autoStore
    @withClientUpdate
    def setFilename(self, f):
        """
        You need to add a docstring here ! 
        """
        self.filename = f


    def getType(self):
        """
        You need to add a docstring here ! 
        """
        return self.type


    @autoStore
    @withClientUpdate
    def setType(self, t):
        """
        You need to add a docstring here ! 
        """
        self.type = t


    observe_setFilename = setFilename
    observe_setType = setType


    def create(self, name):
        """
        You need to add a docstring here ! 
        """
        if self.server == 0:
            return 0
        i = self.server.GetRenderer()
        if self.type == "Mesh":
            return i.createEntity(str(name), self.filename)
        if self.type == "Particles":
            return i.createParticles(str(name), self.filename)
        return 0


    getStateToCacheAndObserveFor = Cacheable.getStateToCacheAndObserveFor
    setCopyableState = Cacheable.setCopyableState



class LocalOgreMeshAsset(MeshAsset):
    """
    An asset that is an ogre mesh existing
    on our local filesystem
    """
    datastoreAttributes = attributes(MeshAsset, filename=text)
    datastoreHash = ("aid", classHash)
    datastoreVersion = 1


    def __init__(self, filename=None, datastore=None):
        MeshAsset.__init__(self)
        self.filename = filename
        self.datastore = datastore


    def getFilename(self):
        """
        You need to add a docstring here ! 
        """
        return self.filename


    @autoStore
    @withClientUpdate
    def setFilename(self, f):
        """
        You need to add a docstring here ! 
        """
        self.filename = f


    observe_setFilename = setFilename


    def create(self, name):
        """
        You need to add a docstring here ! 
        """
        assert self.server, "got no server!"
        i = self.server.GetRenderer()
        return i.createEntity(str(name), self.filename)


    def getStateToCacheAndObserveFor(self, perspective, observer):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.getStateToCacheAndObserveFor(self, perspective, observer)


    def setCopyableState(self, state):
        """
        You need to add a docstring here ! 
        """
        return Cacheable.setCopyableState(self, state)
