# -*- test-case-name: mv3d.test.test_modifier -*-
# Copyright (C) 2008 Mortal Coil Games
# See LICENSE for details.
"""
"""

from zope.interface import Interface, implements

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks
from twisted.spread import pb

from mv3d.net.security import requirePermissions
from mv3d.util.classgen import ClassGenerator
from mv3d.util.conductor import IAssetClient


class IModifiable(Interface):
    """
    Specifies a means of getting modifier interfaces
    """
    def getModifiers(type=None):
        """
        Return the modifiers for a given type, or the default type
        if not specified.
        """


    def addModifier(type, assetid):
        """
        Add a modifier to the list for the given type of
        modification interface
        """



class Modifiable:
    """
    A simple implementation of an IModifiable
    """

    implements(IModifiable)

    modifiers = None
    defaultModifierType = ""

    def getModifiers(self, type=None):
        """
        Return the modifiers for a given type, or the default type
        if not specified
        """
        if self.modifiers is None:
            self.modifiers = dict()
        type = type or self.defaultModifierType
        mods = self.modifiers.get(type, [])
        d = []
        assetService = self.parent.getLocalService(IAssetClient)
        for mod in mods:
            d.append(defer.maybeDeferred(ClassGenerator(assetid=mod,
                assetService=assetService).construct, self))
        return defer.gatherResults(d)


    @requirePermissions("modify")
    def view_getModifiers(self, type=None):
        """
        Return the modifiers for a given type, or the default type
        if not specified
        """
        return self.getModifiers(type)


    def addModifier(self, type, assetid):
        """
        Add a modifier to the list for the given type of
        modification interface
        """
        if self.modifiers is None:
            self.modifiers = dict()
        if not self.modifiers.has_key(type):
            self.modifiers[type] = [assetid]
            return
        self.modifiers[type].append(assetid)



class IModifier(Interface):
    """
    A method of modifying another class
    """



class BoxCreator(pb.Viewable):
    """
    A sample modifier that creates boxes
    """
    implements(IModifier)

    name = "Create Box"

    def __init__(self, parent):
        self.parent = parent


    @inlineCallbacks
    def view_createBox(self, client, position):
        """
        Just create a box
        """
        a = yield self.parent.getItem((0,0))
        r = a.realm
        o = yield a.newObject(ClassGenerator(
            "mv3d.server.model.physical.PhysicalBox"))
        o.create(r)
        o.setName(u"Box%s" % str(position))
        o.setPosition(position)
        o.addVisualAsset((0,17))
        o.body.vobs[0].setScale((0.05, 0.05, 0.05))
        a.takeOut(o.getID())
        a.contain(o.getID())


    def view_getUI(self, client):
        """
        Return the class generator for the UI component
        to this.
        """
        return ClassGenerator(
            "mv3d.client.ui.editor.BoxCreatorView")



